
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>
#include <SDL/SDL_keysym.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "txt.h"
#include "particles.h"
#include "../common/billboard.h"
#include "../common/random.h"


static int _particles_sort(const void *a, const void *b);
static int _particles_draw(struct particle **pl, int n);
static int _particles_compute_face_area(struct particle_system *pa);
static int _particles_random_point_on_face(struct particle_system *pa);


extern struct game_object player_object;
extern struct vector player_ellipsoid;
extern char world_name[256];
extern int world_x, world_y, world_z;


/* the active particle systems list */
struct particle_system *particle_systems = NULL;

/* the world particle systems map */
struct particle_system **particle_system_map = NULL;
struct particle_system **particle_system_table = NULL;

/* particle system */
int particle_systems_id = 0;

/* particle sorting list */
struct particle **particle_list = NULL;
int particle_list_size = 0, particle_list_i = 0;

/* particle texture */
int particle_txt_id;



int particles_init(void) {

  int i;


  /* load the particle texture */
  txt_load("particle-a.s3b", &particle_txt_id);

  particle_system_map = malloc(sizeof(struct particle_system *) * world_x*world_y*world_z);
  particle_system_table = malloc(sizeof(struct game_object *) * MAX_ID);
  if (particle_system_map == NULL || particle_system_table == NULL) {
    fprintf(stderr, "PARTICLE_SYSTEMS_INIT: Out of memory error.\n");
    return FAILED;
  }

  for (i = 0; i < world_x*world_y*world_z; i++)
    particle_system_map[i] = NULL;

  for (i = 0; i < MAX_ID; i++)
    particle_system_table[i] = NULL;

  return SUCCEEDED;
}


int particle_system_get(int id, struct particle_system **ps) {

  if (id < 0 || id > MAX_ID) {
    fprintf(stderr, "PARTICLE_SYSTEM_GET: ID %d is out of range.\n", id);
    *ps = NULL;
    return FAILED;
  }

  *ps = particle_system_table[id];
  if (*ps == NULL) {
    return FAILED;
  }

  return SUCCEEDED;
}


int particle_system_relocate(struct particle_system *ps) {

  unsigned int x, y, z, i;


  x = ps->x / CUBE_SIDE;
  y = ps->y / CUBE_SIDE;
  z = ps->z / CUBE_SIDE;

  if (x < 0 || x >= world_x)
    x = 0;
  if (y < 0 || y >= world_y)
    y = 0;
  if (z < 0 || z >= world_z)
    z = 0;

  /* is the object still in the same cube? */
  if (ps->cx == x && ps->cy == y && ps->cz == z)
    return FAILED;

  /* cube changed -> relocate the object */
  if (ps->next != NULL)
    ps->next->prev = ps->prev;
  if (ps->prev != NULL)
    ps->prev->next = ps->next;
  else {
    i = ps->cz*world_y*world_x + ps->cy*world_x + ps->cx;
    particle_system_map[i] = ps->next;
  }

  ps->cx = x;
  ps->cy = y;
  ps->cz = z;    

  i = z*world_y*world_x + y*world_x + x;
  ps->prev = NULL;
  ps->next = particle_system_map[i];
  if (particle_system_map[i] != NULL)
    particle_system_map[i]->prev = ps;
  particle_system_map[i] = ps;

  return SUCCEEDED;
}


int particle_system_add(int id, int type, float x, float y, float z, float dx, float dy, float dz) {

  struct particle_system *ps;
  unsigned int cx, cy, cz, i;
  double d;
  int a;


  if (particle_system_table[id] != NULL) {
    fprintf(stderr, "PARTICLE_SYSTEM_ADD: ID %d is already in use.\n", id);
    return FAILED;
  }

  ps = malloc(sizeof(struct particle_system));
  if (ps == NULL) {
    fprintf(stderr, "PARTICLE_SYSTEM_ADD: Out of memory error.\n");
    return FAILED;
  }

  /* normalize the direction vector */
  d = sqrt(dx*dx + dy*dy + dz*dz);
  dx /= d;
  dy /= d;
  dz /= d;

  ps->type = type;
  ps->x = x;
  ps->y = y;
  ps->z = z;
  ps->dx = dx;
  ps->dy = dy;
  ps->dz = dz;
  ps->tri = NULL;
  ps->prev = NULL;
  ps->id = id;

  if (ps->tri != NULL)
    _particles_compute_face_area(ps);

  /* amount of max particles */
  if (type == PARTICLE_SYSTEM_POINT_FLAME)
    ps->amount = PARTICLES_MAX;
  else if (type == PARTICLE_SYSTEM_POINT_BLOOD)
    ps->amount = PARTICLES_MAX;
  else if (type == PARTICLE_SYSTEM_FACE_MIST) {
    /* amount = c*area */
    ps->amount = 1024*1*ps->tri[4*3];
  }
  else {
    fprintf(stderr, "PARTICLE_SYSTEM_ADD: Unknown particle system type \"%d\".\n", type);
    return FAILED;
  }

  ps->particles = malloc(sizeof(struct particle) * ps->amount);
  if (ps->particles == NULL) {
    fprintf(stderr, "PARTICLE_SYSTEM_ADD: Out of memory error.\n");
    free(ps);
    return FAILED;
  }

  /* clear the particles */
  for (a = 0; a < ps->amount; a++)
    ps->particles[a].age = 0;

  /* insert the particle system into the map */
  cx = ps->x / CUBE_SIDE;
  cy = ps->y / CUBE_SIDE;
  cz = ps->z / CUBE_SIDE;

  if (cx < 0 || cx >= world_x)
    cx = 0;
  if (cy < 0 || cy >= world_y)
    cy = 0;
  if (cz < 0 || cz >= world_z)
    cz = 0;

  ps->cx = cx;
  ps->cy = cy;
  ps->cz = cz;    

  i = cz*world_y*world_x + cy*world_x + cx;
  ps->prev = NULL;
  ps->next = particle_system_map[i];
  if (particle_system_map[i] != NULL)
    particle_system_map[i]->prev = ps;
  particle_system_map[i] = ps;

  particle_system_table[id] = ps;

  return SUCCEEDED;
}


static int _particles_sort(const void *a, const void *b) {

  struct particle *c = *(struct particle **)a, *d = *(struct particle **)b;

  if (c->d > d->d)
    return -1;
  return 1;
}


static int _particles_draw(struct particle **pl, int n) {

  GLfloat c[4*3], bbv[3*4], texcoords[4*2];
  float a, sc;
  int i;


  billboard_spherical_cheat_matrix(bbv);

  texcoords[0*2 + 0] = 1;
  texcoords[0*2 + 1] = 1;
  texcoords[1*2 + 0] = 0;
  texcoords[1*2 + 1] = 1;
  texcoords[2*2 + 0] = 0;
  texcoords[2*2 + 1] = 0;
  texcoords[3*2 + 0] = 1;
  texcoords[3*2 + 1] = 0;

  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glEnableClientState(GL_VERTEX_ARRAY);

  glTexCoordPointer(2, GL_FLOAT, 0, texcoords);
  glVertexPointer(3, GL_FLOAT, 0, c);

  for (i = 0; i < n; i++) {
    if (pl[i]->age > 0) {
      sc = pl[i]->sc;
      c[2*3 + 0] = pl[i]->x + bbv[0]*sc;
      c[2*3 + 1] = pl[i]->y + bbv[1]*sc;
      c[2*3 + 2] = pl[i]->z + bbv[2]*sc;
      c[3*3 + 0] = pl[i]->x + bbv[3]*sc;
      c[3*3 + 1] = pl[i]->y + bbv[4]*sc;
      c[3*3 + 2] = pl[i]->z + bbv[5]*sc;
      c[0*3 + 0] = pl[i]->x + bbv[6]*sc;
      c[0*3 + 1] = pl[i]->y + bbv[7]*sc;
      c[0*3 + 2] = pl[i]->z + bbv[8]*sc;
      c[1*3 + 0] = pl[i]->x + bbv[9]*sc;
      c[1*3 + 1] = pl[i]->y + bbv[10]*sc;
      c[1*3 + 2] = pl[i]->z + bbv[11]*sc;

      a = (50.0-pl[i]->age)/100;
      if (a < 0) {
	pl[i]->age = 0;
	continue;
      }

      glColor4f(pl[i]->cr, pl[i]->cg, pl[i]->cb, a);

      glDrawArrays(GL_QUADS, 0, 4);
    }
  }

  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);

  return SUCCEEDED;
}


static int _particles_compute_face_area(struct particle_system *pa) {

  float *f, d, e, dx, dy, dz, ex, ey, ez, a;


  f = pa->tri;

  /* compute the area of the triangle */
  dx = f[0*3 + 0] - f[1*3 + 0];
  dy = f[0*3 + 1] - f[1*3 + 1];
  dz = f[0*3 + 2] - f[1*3 + 2];
  d = sqrt(dx*dx + dy*dy + dz*dz);
  ex = f[0*3 + 0] - f[2*3 + 0];
  ey = f[0*3 + 1] - f[2*3 + 1];
  ez = f[0*3 + 2] - f[2*3 + 2];
  e = sqrt(ex*ex + ey*ey + ez*ez);

  a = acos((dx*ex + dy*ey + dz*ez)/(d*e));
  f[4*3] = sin(a)*0.5*d*e;

  return SUCCEEDED;
}


static int _particles_random_point_on_face(struct particle_system *pa) {

  double d[3*3], x, y, z;
  int i;


  for (i = 0; i < 3*3; i++)
    d[i] = pa->tri[i];

  random_point_inside_a_triangle(d, &x, &y, &z);

  pa->tri[3*3 + 0] = x;
  pa->tri[3*3 + 1] = y;
  pa->tri[3*3 + 2] = z;

  return SUCCEEDED;
}


int particles_draw(void) {

  struct txt_object *txt;
  GLint bl, li, te, dm, ff, bd, bs;


  if (particle_list_i == 0)
    return SUCCEEDED;

  txt_get(particle_txt_id, &txt);
  glBindTexture(GL_TEXTURE_2D, txt->ogl_name);

  glGetIntegerv(GL_BLEND, &bl);
  glGetIntegerv(GL_FRONT_FACE, &ff);
  glGetIntegerv(GL_LIGHTING, &li);
  glGetIntegerv(GL_TEXTURE_2D, &te);
  glGetIntegerv(GL_DEPTH_WRITEMASK, &dm);
  glGetIntegerv(GL_BLEND_DST, &bd);
  glGetIntegerv(GL_BLEND_SRC, &bs);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glDepthMask(GL_FALSE);
  glFrontFace(GL_CCW);

  glDisableClientState(GL_NORMAL_ARRAY);
  _particles_draw(particle_list, particle_list_i);
  glEnableClientState(GL_NORMAL_ARRAY);

  /* return old opengl status */
  glBlendFunc(bs, bd);
  glDepthMask(dm);
  glFrontFace(ff);
  if (bl == GL_FALSE)
    glDisable(GL_BLEND);
  if (li == GL_TRUE)
    glEnable(GL_LIGHTING);
  if (te == GL_FALSE)
    glDisable(GL_TEXTURE_2D);

  return SUCCEEDED;
}


int particles_get(int id, struct particle_system **pa) {

  struct particle_system *p;


  p = particle_systems;
  while (p != NULL) {
    if (p->id == id) {
      *pa = p;
      return SUCCEEDED;
    }
  }

  return FAILED;
}


int particles_remove(int id) {

  struct particle_system *p;
  int a, b;


  p = particle_systems;
  while (p != NULL) {
    if (p->id == id) {
      /* free the particle system and reinit the particle list */
      particle_list_size -= p->amount;
      free(particle_list);

      if (p->particles != NULL)
	free(p->particles);

      if (particle_list_size == 0) {
	free(p);
	particle_list = NULL;
	particle_systems = NULL;
	return SUCCEEDED;
      }

      particle_list = malloc(sizeof(struct particle *) * particle_list_size);
      if (particle_list == NULL) {
	particle_list_size = 0;
	fprintf(stderr, "PARTICLES_REMOVE: Out of memory error.\n");
	return FAILED;
      }

      /* remove the system from the system list */
      if (p->prev != NULL)
	p->prev->next = p->next;
      if (p->next != NULL)
	p->next->prev = p->prev;
      if (particle_systems == p)
	particle_systems = p->next;

      free(p);

      /* reinit the particle list */
      p = particle_systems;
      b = 0;
      while (p != NULL) {
	for (a = 0; a < p->amount; a++, b++)
	  particle_list[b] = &(p->particles[a]);
	p = p->next;
      }

      return SUCCEEDED;
    }
    p = p->next;
  }

  return FAILED;
}


int particles_add(int type, float x, float y, float z, float dx, float dy, float dz, float *tri, int *id) {

  struct particle_system *p;
  float d;
  int a, b;


  p = malloc(sizeof(struct particle_system));
  if (p == NULL) {
    fprintf(stderr, "PARTICLES_ADD: Out of memory error.\n");
    return FAILED;
  }

  /* normalize the direction vector */
  d = sqrt(dx*dx + dy*dy + dz*dz);
  dx /= d;
  dy /= d;
  dz /= d;

  p->type = type;
  p->x = x;
  p->y = y;
  p->z = z;
  p->dx = dx;
  p->dy = dy;
  p->dz = dz;
  p->tri = tri;
  p->prev = NULL;
  p->id = particle_systems_id++;
  *id = p->id;

  if (tri != NULL)
    _particles_compute_face_area(p);

  /* amount of max particles */
  if (type == PARTICLE_SYSTEM_POINT_FLAME)
    p->amount = PARTICLES_MAX;
  else if (type == PARTICLE_SYSTEM_POINT_BLOOD)
    p->amount = PARTICLES_MAX;
  else if (type == PARTICLE_SYSTEM_FACE_MIST) {
    /* amount = c*area */
    p->amount = 1024*1*p->tri[4*3];
  }
  else {
    fprintf(stderr, "PARTICLES_ADD: Unknown particle system type \"%d\".\n", type);
    return FAILED;
  }

  p->particles = malloc(sizeof(struct particle) * p->amount);
  if (p->particles == NULL) {
    fprintf(stderr, "PARTICLES_ADD: Out of memory error.\n");
    return FAILED;
  }

  /* clear the particles */
  for (a = 0; a < p->amount; a++)
    p->particles[a].age = 0;

  /* add the new system to the list of systems */
  p->next = particle_systems;
  if (particle_systems != NULL)
    particle_systems->prev = p;
  particle_systems = p;

  /* add the new particles to the particle sort list */
  if (particle_list == NULL)
    particle_list = malloc(sizeof(struct particle *) * p->amount);
  else
    particle_list = realloc(particle_list, sizeof(struct particle *) * (particle_list_size + p->amount));

  if (particle_list == NULL) {
    fprintf(stderr, "PARTICLES_ADD: Out of memory error.\n");
    return FAILED;
  }

  for (a = 0, b = particle_list_size; a < p->amount; a++, b++)
    particle_list[b] = &(p->particles[a]);

  particle_list_size += p->amount;

  return SUCCEEDED;
}


int particles_update(void) {

  struct particle_system *ps;
  struct particle *p;
  double dx, dy, dz, r;
  int i, d, z, j, k;


  /* recreate the particle buffer */
  ps = particle_systems;
  j = 0;
  while (ps != NULL) {
    j += ps->amount;
    ps = ps->next_ext;
  }

  if (particle_list_size < j) {
    particle_list_size = j;
    particle_list = realloc(particle_list, sizeof(struct particle *) * particle_list_size);
    if (particle_list == NULL) {
      fprintf(stderr, "DRAW_MAIN_GLAREA: Out of memory error.\n");
      return FAILED;
    }
  }

  particle_list_i = j;

  ps = particle_systems;
  k = 0;
  while (ps != NULL) {
    for (j = 0; j < ps->amount; j++, k++)
      particle_list[k] = &(ps->particles[j]);
    ps = ps->next_ext;
  }

  ps = particle_systems;
  if (ps == NULL)
    return SUCCEEDED;

  while (ps != NULL) {
    p = ps->particles;

    /* count the dead */
    for (d = 0, i = 0; i < ps->amount; i++) {
      if (p[i].age == 0)
	d++;
    }

    /* POINT BLOOD */
    if (ps->type == PARTICLE_SYSTEM_POINT_BLOOD) {

      /* wake up a few */
      if (d > 0) {
	for (z = 0, i = 0; i < ps->amount; i++) {
	  if (p[i].age == 0) {

	    /* compute a random direction */
	    dx = rand()/(double)RAND_MAX-0.5;
	    dy = rand()/(double)RAND_MAX-0.5;
	    dz = rand()/(double)RAND_MAX-0.5;
	    r = sqrt(dx*dx + dy*dy + dz*dz);
	    dx /= r;
	    dy /= r;
	    dz /= r;
	    r = dx*ps->dx + dy*ps->dy + dz*ps->dz;
	    dx = (dx - r*ps->dx)*0.002 + ps->dx*0.01;
	    dy = (dy - r*ps->dy)*0.002 + ps->dy*0.01;
	    dz = (dz - r*ps->dz)*0.002 + ps->dz*0.01;

	    p[i].age = 1;
	    p[i].x = ps->x;
	    p[i].y = ps->y;
	    p[i].z = ps->z;
	    p[i].vx = dx;
	    p[i].vy = dy;
	    p[i].vz = dz;
	    p[i].ax = 0;
	    p[i].ay = -0.002;
	    p[i].az = 0;
	    p[i].cr = 0.5;
	    p[i].cg = 0;
	    p[i].cb = 0;
	    p[i].sc = 0.02;
	    z++;
	    d--;
	    if (d == 0)
	      break;
	    if (z == 10)
	      break;
	  }
	}
      }
    }
    /* POINT FLAME */
    else if (ps->type == PARTICLE_SYSTEM_POINT_FLAME) {

      /* wake up a few */
      if (d > 0) {
	for (z = 0, i = 0; i < ps->amount; i++) {
	  if (p[i].age == 0) {

	    /* compute a random direction */
	    dx = rand()/(double)RAND_MAX-0.5;
	    dy = rand()/(double)RAND_MAX-0.5;
	    dz = rand()/(double)RAND_MAX-0.5;
	    r = sqrt(dx*dx + dy*dy + dz*dz);
	    dx /= r;
	    dy /= r;
	    dz /= r;
	    r = dx*ps->dx + dy*ps->dy + dz*ps->dz;
	    dx = (dx - r*ps->dx)*0.0006 + ps->dx*0.001;
	    dy = (dy - r*ps->dy)*0.0006 + ps->dy*0.001;
	    dz = (dz - r*ps->dz)*0.0006 + ps->dz*0.001;

	    p[i].age = 1;
	    p[i].x = ps->x;
	    p[i].y = ps->y;
	    p[i].z = ps->z;
	    p[i].vx = dx;
	    p[i].vy = dy;
	    p[i].vz = dz;
	    p[i].ax = ps->dx * 0.0002;
	    p[i].ay = ps->dy * 0.0002;
	    p[i].az = ps->dz * 0.0002;
	    p[i].cr = 0.4;
	    p[i].cg = 0;
	    p[i].cb = 0.5;
	    if ((rand() % 10) == 0) {
	      p[i].cr = 1;
	      p[i].cg = 1;
	      p[i].cb = 1;
	    }
	    p[i].sc = 0.01;
	    z++;
	    d--;
	    if (d == 0)
	      break;
	    if (z == 10)
	      break;
	  }
	}
      }
    }
    /* FACE MIST */
    else if (ps->type == PARTICLE_SYSTEM_FACE_MIST) {

      /* wake up a few */
      if (d > 0) {
	for (z = 0, i = 0; i < ps->amount; i++) {
	  if (p[i].age == 0) {

	    _particles_random_point_on_face(ps);

	    p[i].age = 1;
	    p[i].x = ps->tri[3*3 + 0];
	    p[i].y = ps->tri[3*3 + 1];
	    p[i].z = ps->tri[3*3 + 2];
	    p[i].vx = 0;
	    p[i].vy = 0;
	    p[i].vz = 0;
	    p[i].ax = ps->dx * 0.0001;
	    p[i].ay = ps->dy * 0.0001;
	    p[i].az = ps->dz * 0.0001;
	    p[i].cr = 1;
	    p[i].cg = 1;
	    p[i].cb = 0;
	    p[i].sc = 0.005;
	    z++;
	    d--;
	    if (d == 0)
	      break;
	    if (z >= (int)(20*1*ps->tri[4*3]))
	      break;
	  }
	}
      }
    }

    /* advance the living particles */
    for (i = 0; i < ps->amount; i++) {
      if (p[i].age > 0) {
	p[i].age += 1;
	p[i].x += p[i].vx;
	p[i].y += p[i].vy;
	p[i].z += p[i].vz;
	p[i].vx += p[i].ax;
	p[i].vy += p[i].ay;
	p[i].vz += p[i].az;
	/* compute particle distance to the viewer */
	dx = player_object.x - p[i].x;
	dy = player_object.y + player_ellipsoid.y/2 - p[i].y;
	dz = player_object.z - p[i].z;
	p[i].d = dx*dx + dy*dy + dz*dz;
      }
    }

    ps = ps->next_ext;
  }

  /* sort the particles (distance to the camera) */
  qsort(particle_list, particle_list_i, sizeof(struct particle *), _particles_sort);

  return SUCCEEDED;
}
