
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "defines.h"
#include "mesh2.h"
#include "parse.h"
#include "../client/txt.h"


int mesh2_create(struct mesh2 **m2, int tri, int ver, int obj, int display_errors) {

  char memerr[] = "MESH2_CREATE: Out of memory error.\n";
  struct mesh2 *m;


  m = malloc(sizeof(struct mesh2));
  if (m == NULL) {
    if (display_errors == YES)
      fprintf(stderr, memerr);
    return FAILED;
  }

  m->triangles = tri;
  m->vertices = ver;
  m->objects = obj;

  m->vertex_list = malloc(sizeof(float)*3*3*tri);
  m->vertex_normal_list = malloc(sizeof(float)*3*3*tri);
  m->face_normal_list = malloc(sizeof(float)*3*tri);
  m->uv_list = malloc(sizeof(float)*2*3*tri);
  m->triangle_radius_list = malloc(sizeof(float)*tri);
  if (m->vertex_list == NULL || m->vertex_normal_list == NULL ||
      m->face_normal_list == NULL || m->uv_list == NULL ||
      m->triangle_radius_list == NULL) {
    free(m);
    if (display_errors == YES)
      fprintf(stderr, memerr);
    return FAILED;
  }

  *m2 = m;

  return SUCCEEDED;
}


int mesh2_free(struct mesh2 *m) {

  if (m == NULL)
    return FAILED;

  if (m->vertex_list != NULL)
    free(m->vertex_list);
  if (m->vertex_normal_list != NULL)
    free(m->vertex_normal_list);
  if (m->face_normal_list != NULL)
    free(m->face_normal_list);
  if (m->uv_list != NULL)
    free(m->uv_list);
  if (m->triangle_radius_list != NULL)
    free(m->triangle_radius_list);
  free(m);

  return SUCCEEDED;
}


int mesh2_compute_bounding_sphere(struct mesh2 *mesh, float *sphere) {

  float nx, ny, nz, px, py, pz, mx, my, mz, *t;
  int i;


  t = mesh->vertex_list;
  nx = *(t++);
  px = nx;
  ny = *(t++);
  py = ny;
  nz = *(t++);
  pz = nz;

  /* first compute the bounding box */
  for (i = 1; i < mesh->vertices; i++) {
    if (*(t+0) < nx)
      nx = *(t+0);
    else if (*(t+0) > px)
      px = *(t+0);
    if (*(t+1) < ny)
      ny = *(t+1);
    else if (*(t+1) > py)
      py = *(t+1);
    if (*(t+2) < nz)
      nz = *(t+2);
    else if (*(t+2) > pz)
      pz = *(t+2);
    t += 3;
  }

  /* then create the bouding sphere surrounding the bounding box */
  mx = (px-nx)/2;
  my = (py-ny)/2;
  mz = (pz-nz)/2;

  /* write the sphere radius and centre */
  sphere[0] = mx+nx;
  sphere[1] = my+ny;
  sphere[2] = mz+nz;
  sphere[4] = mx*mx + my*my + mz*mz;
  sphere[3] = sqrt(sphere[4]);

  return SUCCEEDED;
}


int mesh2_compute_triangle_bounding_spheres(struct mesh2 *mesh, int sqroot) {

  float dx1, dx2, dy1, dy2, dz1, dz2, x1, x2, *f, *t;
  int i;


  if (mesh == NULL)
    return FAILED;

  if (mesh->triangle_radius_list == NULL) {
    mesh->triangle_radius_list = malloc(sizeof(float) * mesh->triangles);
    if (mesh->triangle_radius_list == NULL) {
      fprintf(stderr, "MESH2_COMPUTE_TRIANGLE_BOUNDING_SPHERES: Out of memory error.\n");
      return FAILED;
    }
  }

  /* estimate triangles with spheres, compute radius for such a sphere that encloses the whole triangle */
  f = mesh->triangle_radius_list;
  t = mesh->vertex_list;
  for (i = 0; i < mesh->triangles; i++, t += 9) {
    dx1 = t[0] - t[3];
    dy1 = t[1] - t[4];
    dz1 = t[2] - t[5];
    dx2 = t[0] - t[6];
    dy2 = t[1] - t[7];
    dz2 = t[2] - t[8];

    x1 = dx1*dx1 + dy1*dy1 + dz1*dz1;
    x2 = dx2*dx2 + dy2*dy2 + dz2*dz2;

    if (sqroot == NO) {
      if (x1 >= x2)
	*(f++) = x1;
      else
	*(f++) = x2;
    }
    else {
      if (x1 >= x2)
	*(f++) = sqrt(x1);
      else
	*(f++) = sqrt(x2);
    }
  }

  return SUCCEEDED;
}


int mesh2_read_ascii(char *name, struct mesh2 **mesh, int display_errors) {

  char b[64];
  struct mesh2 *m;
  FILE *f;
  int tri, ver, obj, i;


  f = fopen(name, "rb");
  if (f == NULL) {
    if (display_errors == YES)
      fprintf(stderr, "MESH2_READ_ASCII: Could not open file \"%s\".\n", name);
    return FAILED;
  }

  /* read the header */
  fscanf(f, "%63s", b);
  if (strcmp(b, "MESH2") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"%s\" is not a MESH2 file.\n", name);
    fclose(f);
    return FAILED;
  }

  /* triangles */
  fscanf(f, "%63s", b);
  if (strcmp(b, "Triangles:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"Triangles:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  fscanf(f, "%d", &tri);

  /* vertices */
  fscanf(f, "%63s", b);
  if (strcmp(b, "Vertices:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"Vertices:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  fscanf(f, "%d", &ver);

  /* objects */
  fscanf(f, "%63s", b);
  if (strcmp(b, "Objects:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"Objects:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  fscanf(f, "%d", &obj);

  /* create the object */
  if (mesh2_create(&m, tri, ver, obj, display_errors) == FAILED) {
    fclose(f);
    return FAILED;
  }

  /* object data */
  fscanf(f, "%63s", b);
  if (strcmp(b, "ObjectData:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"ObjectData:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  for (i = 0; i < obj; i++) {
    fscanf(f, "%d %d", &(m->object_data[i*2 + 0]), &(m->object_data[i*2 + 1]));
  }

  /* texture data */
  fscanf(f, "%63s", b);
  if (strcmp(b, "TextureData:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"TextureData:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  /* read the textures and convert the id */
  for (i = 0; i < obj; i++) {
    fscanf(f, "%63s", b);
    b[strlen(b)-1] = 0;
    txt_load(b+1, &(m->object_data[i*2 + 0]));
  }

  /* vertex data */
  fscanf(f, "%63s", b);
  if (strcmp(b, "VertexData:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"VertexData:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  for (i = 0; i < ver; i++) {
    fscanf(f, "%f %f %f", &(m->vertex_list[i*3 + 0]), &(m->vertex_list[i*3 + 1]), &(m->vertex_list[i*3 + 2]));
  }

  /* vertex normal data */
  fscanf(f, "%63s", b);
  if (strcmp(b, "VertexNormalData:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"VertexNormalData:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  for (i = 0; i < ver; i++) {
    fscanf(f, "%f %f %f", &(m->vertex_normal_list[i*3 + 0]), &(m->vertex_normal_list[i*3 + 1]), &(m->vertex_normal_list[i*3 + 2]));
  }

  /* face normal data */
  fscanf(f, "%63s", b);
  if (strcmp(b, "FaceNormalData:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"FaceNormalData:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  for (i = 0; i < tri; i++) {
    fscanf(f, "%f %f %f", &(m->face_normal_list[i*3 + 0]), &(m->face_normal_list[i*3 + 1]), &(m->face_normal_list[i*3 + 2]));
  }

  /* texture coordinate data */
  fscanf(f, "%63s", b);
  if (strcmp(b, "TexCoordData:") != 0) {
    fprintf(stderr, "MESH2_READ_ASCII: \"TexCoordData:\" not found in \"%s\".\n", name);
    fclose(f);
    return FAILED;
  }

  for (i = 0; i < tri; i++) {
    fscanf(f, "%f %f", &(m->uv_list[i*3*2 + 0]), &(m->uv_list[i*3*2 + 1]));
    fscanf(f, "%f %f", &(m->uv_list[i*3*2 + 2]), &(m->uv_list[i*3*2 + 3]));
    fscanf(f, "%f %f", &(m->uv_list[i*3*2 + 4]), &(m->uv_list[i*3*2 + 5]));
  }

  fclose(f);

  *mesh = m;

  return SUCCEEDED;
}


int mesh2_write_ascii(struct mesh2 *m) {

  struct txt_object *t;
  int i;


  printf("MESH2\n");
  printf("Triangles: %d\n", m->triangles);
  printf("Vertices: %d\n", m->vertices);
  printf("Objects: %d\n", m->objects);

  /* object data */
  printf("ObjectData:\n");

  for (i = 0; i < m->objects; i++) {
    printf("%d %d\n", m->object_data[i*2 + 0], m->object_data[i*2 + 1]);
  }

  /* texture data */
  printf("TextureData:\n");

  for (i = 0; i < m->objects; i++) {
    txt_get(i, &t);
    printf("\"%s\"\n", t->name);
  }

  /* vertex data */
  printf("VertexData:\n");

  for (i = 0; i < m->vertices; i++) {
    printf("%f %f %f\n", m->vertex_list[i*3 + 0], m->vertex_list[i*3 + 1], m->vertex_list[i*3 + 2]);
  }

  /* vertex normals */
  printf("VertexNormalData:\n");

  for (i = 0; i < m->vertices; i++) {
    printf("%f %f %f\n", m->vertex_normal_list[i*3 + 0], m->vertex_normal_list[i*3 + 1], m->vertex_normal_list[i*3 + 2]);
  }

  /* face normals */
  printf("FaceNormalData:\n");

  for (i = 0; i < m->triangles; i++) {
    printf("%f %f %f\n", m->face_normal_list[i*3 + 0], m->face_normal_list[i*3 + 1], m->face_normal_list[i*3 + 2]);
  }

  /* texture coordinate data */
  printf("TexCoordData:\n");

  for (i = 0; i < m->triangles; i++) {
    printf("%f %f ", m->uv_list[i*3*2 + 0], m->uv_list[i*3*2 + 1]);
    printf("%f %f ", m->uv_list[i*3*2 + 2], m->uv_list[i*3*2 + 3]);
    printf("%f %f\n", m->uv_list[i*3*2 + 4], m->uv_list[i*3*2 + 5]);
  }

  return SUCCEEDED;
}


int mesh2_read_binary(char *name, struct mesh2 **mesh, int display_errors) {

  struct mesh2 *m;
  int tri = 0, ver = 0, obj = 0, i, n;
  char tmp[64];
  FILE *f;


  f = fopen(name, "rb");
  if (f == NULL) {
    if (display_errors == YES)
      fprintf(stderr, "MESH2_READ_BINARY: Could not open file \"%s\".\n", name);
    return FAILED;
  }

  fscanf(f, "%c%c%c%c%c", &tmp[0], &tmp[1], &tmp[2], &tmp[3], &tmp[4]);
  tmp[5] = 0;
  if (strcmp("MESH2", tmp) != 0) {
    if (display_errors == YES)
      fprintf(stderr, "MESH2_READ_BINARY: \"%s\" is not a MESH2 file.\n", name);
    fclose(f);
    return FAILED;
  }

  file_read_int(f, &tri);
  file_read_int(f, &ver);
  file_read_int(f, &obj);

  /* create the object */
  if (mesh2_create(&m, tri, ver, obj, display_errors) == FAILED) {
    fclose(f);
    return FAILED;
  }

  /* object data */
  for (i = 0; i < m->objects; i++) {
    file_read_int(f, &m->object_data[i*2 + 0]);
    file_read_int(f, &m->object_data[i*2 + 1]);
  }

  /* texture data */
  for (i = 0; i < m->objects; i++) {
    for (n = 0; n < 63; n++) {
      fscanf(f, "%c", &tmp[n]);
      if (tmp[n] == 0)
	break;
    }
    txt_load(tmp, &(m->object_data[i*2 + 0]));
  }

  /* vertex data */
  for (i = 0; i < m->vertices; i++) {
    file_read_float(f, &m->vertex_list[i*3 + 0]);
    file_read_float(f, &m->vertex_list[i*3 + 1]);
    file_read_float(f, &m->vertex_list[i*3 + 2]);
  }

  /* vertex normals */
  for (i = 0; i < m->vertices; i++) {
    file_read_float(f, &m->vertex_normal_list[i*3 + 0]);
    file_read_float(f, &m->vertex_normal_list[i*3 + 1]);
    file_read_float(f, &m->vertex_normal_list[i*3 + 2]);
  }

  /* face normals */
  for (i = 0; i < m->triangles; i++) {
    file_read_float(f, &m->face_normal_list[i*3 + 0]);
    file_read_float(f, &m->face_normal_list[i*3 + 1]);
    file_read_float(f, &m->face_normal_list[i*3 + 2]);
  }

  /* texture coordinate data */
  for (i = 0; i < m->triangles; i++) {
    file_read_float(f, &m->uv_list[i*3*2 + 0]);
    file_read_float(f, &m->uv_list[i*3*2 + 1]);
    file_read_float(f, &m->uv_list[i*3*2 + 2]);
    file_read_float(f, &m->uv_list[i*3*2 + 3]);
    file_read_float(f, &m->uv_list[i*3*2 + 4]);
    file_read_float(f, &m->uv_list[i*3*2 + 5]);
  }

  fclose(f);

  *mesh = m;

  return SUCCEEDED;
}


int mesh2_write_binary(char *name, struct mesh2 *m) {

  struct txt_object *t;
  FILE *f;
  int i;


  f = fopen(name, "wb");
  if (f == NULL) {
    fprintf(stderr, "MESH2_WRITE_BINARY: Could not open file \"%s\" for writing.\n", name);
    return FAILED;
  }

  fprintf(f, "MESH2");

  file_write_int(f, m->triangles);
  file_write_int(f, m->vertices);
  file_write_int(f, m->objects);

  /* object data */
  for (i = 0; i < m->objects; i++) {
    file_write_int(f, m->object_data[i*2 + 0]);
    file_write_int(f, m->object_data[i*2 + 1]);
  }

  /* texture data */
  for (i = 0; i < m->objects; i++) {
    txt_get(i, &t);
    fprintf(f, "%s%c", t->name, 0);
  }

  /* vertex data */
  for (i = 0; i < m->vertices; i++) {
    file_write_float(f, m->vertex_list[i*3 + 0]);
    file_write_float(f, m->vertex_list[i*3 + 1]);
    file_write_float(f, m->vertex_list[i*3 + 2]);
  }

  /* vertex normals */
  for (i = 0; i < m->vertices; i++) {
    file_write_float(f, m->vertex_normal_list[i*3 + 0]);
    file_write_float(f, m->vertex_normal_list[i*3 + 1]);
    file_write_float(f, m->vertex_normal_list[i*3 + 2]);
  }

  /* face normals */
  for (i = 0; i < m->triangles; i++) {
    file_write_float(f, m->face_normal_list[i*3 + 0]);
    file_write_float(f, m->face_normal_list[i*3 + 1]);
    file_write_float(f, m->face_normal_list[i*3 + 2]);
  }

  /* texture coordinate data */
  for (i = 0; i < m->triangles; i++) {
    file_write_float(f, m->uv_list[i*3*2 + 0]);
    file_write_float(f, m->uv_list[i*3*2 + 1]);
    file_write_float(f, m->uv_list[i*3*2 + 2]);
    file_write_float(f, m->uv_list[i*3*2 + 3]);
    file_write_float(f, m->uv_list[i*3*2 + 4]);
    file_write_float(f, m->uv_list[i*3*2 + 5]);
  }

  fclose(f);

  return SUCCEEDED;
}
