
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "../common/defines.h"
#include "../common/mesh.h"
#include "txt.h"
#include "simplify.h"

/*
#define _SIMPLIFY_DEBUG 1
*/

#define COST_MAX 10000.0


/* NOTE! i'm using a simplification algorithm my stan melax
   http://www.melax.com/polychop/index.html */


float _min(float a, float b) {

  if (a < b)
    return a;

  return b;
}


float _max(float a, float b) {

  if (a > b)
    return a;

  return b;
}


/* compute the cost for moving vertex a to b */
float _compute_cost(struct object_mesh *m, int a, int b, int *triangles, int t, int *faces, int f, char *tr) {

  float cv, dx, dy, dz, e, mc, *fnl;
  int i, *tri, p1, p2, p3, j;


  /* none or far too many faces sharing this edge -> ultra high cost */
  if (f == 0 || f == 256 || t == 0 || t == 256)
    return COST_MAX;

  /* check that all the faces use the same texture, otherwise
     the cost will be ultra high! */
  i = m->textures[faces[0]/3];
  for (j = 1; j < f; j++) {
    if (m->textures[faces[j]] != i)
      return COST_MAX;
  }

  cv = 0;

  /* compute the cost */
  tri = m->triangle_list;
  fnl = m->face_normal_list;
  for (i = 0; i < t; i++) {
    /* does this triangle contain a? */
    p2 = triangles[i];
    if (a == tri[p2 + 0] || a == tri[p2 + 1] || a == tri[p2 + 2]) {
      mc = 2;
      for (j = 0; j < f; j++) {
	/* take the dot product of the face normals */
	for (p3 = 0; p3 < f; p3++) {
	  if (p2 == faces[p3])
	    break;
	}
	if (p3 == f) {
	  p1 = faces[j];
	  dx = fnl[p2 + 0]*fnl[p1 + 0] + fnl[p2 + 1]*fnl[p1 + 1] + fnl[p2 + 2]*fnl[p1 + 2];
	  mc = _min(mc, 1 - dx);
	}
      }
      cv = _max(cv, mc);
    }
  }

  /* the compute edge length */
  a *= 3;
  b *= 3;
  dx = m->vertice_list[a + 0] - m->vertice_list[b + 0];
  dy = m->vertice_list[a + 1] - m->vertice_list[b + 1];
  dz = m->vertice_list[a + 2] - m->vertice_list[b + 2];
  e = sqrt(dx*dx + dy*dy + dz*dz);

  return e*cv/2;
}


int simplify_mesh(struct object_mesh *m, float quality) {

  float *ct, min, t;
  char *ve, *tr, *co;
  int i, j, k, l, n, mv, *tl, ri, rj, p1, p2, p3, *in, *tri;
  int cj, ck, cl, trianglesj[256], trianglesk[256], trianglesl[256];
  int fjk, fjl, fkl, facesjk[256], facesjl[256], faceskl[256];
  double *vl;


  if (m == NULL)
    return FAILED;

  mv = m->vertices;
  ct = malloc(mv * sizeof(float));
  in = malloc(mv * sizeof(int));
  ve = malloc(mv);
  co = malloc(m->triangles);
  tr = malloc(m->triangles);

  if (ct == NULL || ve == NULL || tr == NULL || in == NULL || co == NULL) {
    if (ct != NULL)
      free(ct);
    if (ve != NULL)
      free(ve);
    if (tr != NULL)
      free(tr);
    if (in != NULL)
      free(in);
    if (co != NULL)
      free(co);
    fprintf(stderr, "%d\n", mv + m->triangles + (mv*mv*sizeof(float)));
    fprintf(stderr, "SIMPLIFY_MESH: Out of memory error.\n");
    return FAILED;
  }

  /* init to: keep all vertices */
  for (i = 0; i < mv; i++)
    ve[i] = 1;

  /* init to: keep all triangles */
  for (i = 0; i < m->triangles; i++)
    tr[i] = 1;

  /* init to: maximum cost for every move */
  for (i = 0; i < mv; i++) {
    ct[i] = COST_MAX;
    in[i] = 0;
  }

  /* init to: no computations done */
  for (i = 0; i < m->triangles; i++)
    co[i] = 0;

  /* collapse edges until the quality threshold is reached */
  tl = m->triangle_list;
  while (1) {

    /* find costs for every move */
    for (i = 0; i < m->triangles; i++) {
      fprintf(stderr, "triangles: %d\n", i);
      if (tr[i] == 1 && co[i] == 0) {
	j = tl[i*3 + 0];
	k = tl[i*3 + 1];
	l = tl[i*3 + 2];

	/* find the faces using these two vertices */
	tri = m->triangle_list;
	for (fjk = 0, fjl = 0, fkl = 0, cj = 0, ck = 0, cl = 0, n = 0; n < m->triangles; n++, tri += 3) {
	  if (tr[n] == 1) {
	    p1 = tri[0];
	    p2 = tri[1];
	    p3 = tri[2];
	    if (p1 == j || p2 == j || p3 == j) {
	      if (p1 == k || p2 == k || p3 == k) {
		facesjk[fjk++] = n*3;
		if (fjk == 256)
		  break;
	      }
	      trianglesj[cj++] = n*3;
	      if (cj == 256)
		break;
	    }
	    if (p1 == k || p2 == k || p3 == k) {
	      if (p1 == l || p2 == l || p3 == l) {
		faceskl[fkl++] = n*3;
		if (fkl == 256)
		  break;
	      }
	      trianglesk[ck++] = n*3;
	      if (ck == 256)
		break;
	    }
	    if (p1 == l || p2 == l || p3 == l) {
	      if (p1 == j || p2 == j || p3 == j) {
		faceskl[fjl++] = n*3;
		if (fjl == 256)
		  break;
	      }
	      trianglesl[cl++] = n*3;
	      if (cl == 256)
		break;
	    }
	  }
	}

	/* move vertex 0 to 1 and 2 */
	t = _compute_cost(m, j, k, trianglesj, cj, facesjk, fjk, tr);
	if (ct[j]> t) {
	  ct[j] = t;
	  in[j] = k;
	}
	t = _compute_cost(m, j, l, trianglesj, cj, facesjl, fjl, tr);
	if (ct[j] > t) {
	  ct[j] = t;
	  in[j] = l;
	}

	/* move vertex 1 to 0 and 2 */
	t = _compute_cost(m, k, j, trianglesk, ck, facesjk, fjk, tr);
	if (ct[k] > t) {
	  ct[k] = t;
	  in[k] = j;
	}
	t = _compute_cost(m, k, l, trianglesk, ck, faceskl, fkl, tr);
	if (ct[k] > t) {
	  ct[k] = t;
	  in[k] = l;
	}

	/* move vertex 2 to 0 and 1 */
	t = _compute_cost(m, l, j, trianglesl, cl, facesjl, fjl, tr);
	if (ct[l] > t) {
	  ct[l] = t;
	  in[l] = j;
	}
	t = _compute_cost(m, l, k, trianglesl, cl, faceskl, fkl, tr);
	if (ct[l] > t) {
	  ct[l] = t;
	  in[l] = k;
	}
	/* mark it computed */
	co[i] = 1;
      }
    }

    /* find the minimum cost */
    min = COST_MAX;
    ri = 0;
    rj = 0;
    for (i = 0; i < mv; i++) {
      if (ve[i] == 1 && ve[in[i]] == 1 && ct[i] < min) {
	ri = i;
	rj = in[i];
	min = ct[i];
      }
    }

    if (ri == rj)
      break;

    /* have we reached the limit? */
    if (min > quality)
      break;

    fprintf(stderr, "MIN: %f\n", min);

#ifdef _SIMPLIFY_DEBUG
    fprintf(stderr, "SIMPLIFY_MESH: Move %d->%d cost %f.\n", ri, rj, min);
#endif

    /* remove the faces that use these two vertices */
    for (i = 0, j = 0; i < m->triangles && j < 256; i++) {
      if (tr[i] == 1) {
	p1 = tl[i*3 + 0];
	p2 = tl[i*3 + 1];
	p3 = tl[i*3 + 2];
	if (((p1 == ri || p1 == rj) && (p2 == ri || p2 == rj)) ||
	    ((p1 == ri || p1 == rj) && (p3 == ri || p3 == rj)) ||
	    ((p2 == ri || p2 == rj) && (p3 == ri || p3 == rj))) {
	  tr[i] = 0;
	}
      }
    }

    /* take the vertex away */
    ve[ri] = 0;

    /* move it into the other one */
    vl = m->vertice_list;
    for (i = 0, k = 0; k < m->triangles; i += 3, k++) {
      if (tr[k] == 1) {
	j = 0;
	if (tl[i + 0] == ri) {
	  tl[i + 0] = rj;
	  j++;
	}
	else if (tl[i + 1] == ri) {
	  tl[i + 1] = rj;
	  j++;
	}
	else if (tl[i + 2] == ri) {
	  tl[i + 2] = rj;
	  j++;
	}
	/* a change? if yes, recompute the face normal and mark it uncomputed */
	if (j > 0) {

	  double x1, y1, z1, x2, y2, z2, x3, y3, z3, dx1, dy1, dz1, dx2, dy2, dz2;
	  int n;

	  n = tl[i + 0]*3;
	  x1 = vl[n + 0];
	  y1 = vl[n + 1];
	  z1 = vl[n + 2];
	  n = tl[i + 1]*3;
	  x2 = vl[n + 0];
	  y2 = vl[n + 1];
	  z2 = vl[n + 2];
	  n = tl[i + 2]*3;
	  x3 = vl[n + 0];
	  y3 = vl[n + 1];
	  z3 = vl[n + 2];

	  dx1 = x2 - x1;
	  dy1 = y2 - y1;
	  dz1 = z2 - z1;
	  dx2 = x3 - x1;
	  dy2 = y3 - y1;
	  dz2 = z3 - z1;
	  x1 = dy1*dz2 - dy2*dz1;
	  x2 = dx2*dz1 - dx1*dz2;
	  x3 = dx1*dy2 - dx2*dy1;

	  dx1 = sqrt(x1*x1 + x2*x2 + x3*x3);

	  m->face_normal_list[i + 0] = (float)(x1/dx1);
	  m->face_normal_list[i + 1] = (float)(x2/dx1);
	  m->face_normal_list[i + 2] = (float)(x3/dx1);

	  co[k] = 0;
	}
      }
    }
  }

  /* init to: keep no vertices */
  for (i = 0; i < m->vertices; i++)
    ve[i] = 0;

  /* keep only those vertices that the triangles use */
  for (i = 0; i < m->triangles; i++) {
    if (tr[i] == 1) {
      ve[tl[i*3 + 0]] = 1;
      ve[tl[i*3 + 1]] = 1;
      ve[tl[i*3 + 2]] = 1;
    }
  }

#ifdef _SIMPLIFY_DEBUG
  /* debug output */
  fprintf(stderr, "TRIANGLES:\n");
  for (i = 0; i < m->triangles; i++)
    fprintf(stderr, "%d ", tr[i]);
  fprintf(stderr, "\n");

  fprintf(stderr, "VERTICES:\n");
  for (i = 0; i < m->vertices; i++)
    fprintf(stderr, "%d ", ve[i]);
  fprintf(stderr, "\n");
#endif

  /* count the remaining triangles and vertices */
  for (i = 0, p1 = 0; i < m->triangles; i++) {
    if (tr[i] == 1)
      p1++;
  }

  for (i = 0, p2 = 0; i < m->vertices; i++) {
    if (ve[i] == 1)
      p2++;
  }

  fprintf(stderr, "SIMPLIFY_MESH: mesh \"%s\", triangles %d -> %d, vertices %d -> %d\n", m->name, m->triangles, p1, m->vertices, p2);

  /* compact the vertices */
  for (i = 0; i < m->vertices; i++) {
    if (ve[i] == 0) {
      for (j = m->vertices-1; j > i && ve[j] == 0; j--);
      if (j > i) {
	ve[j] = 0;
	ve[i] = 1;
	m->vertice_list[i*3 + 0] = m->vertice_list[j*3 + 0];
	m->vertice_list[i*3 + 1] = m->vertice_list[j*3 + 1];
	m->vertice_list[i*3 + 2] = m->vertice_list[j*3 + 2];
	/* update the triangles */
	for (ri = 0; ri < m->triangles*3; ri++) {
	  if (tl[ri] == j)
	    tl[ri] = i;
	}
      }
      else
	i = m->vertices;
    }
  }

  /* compact triangles */
  for (i = 0; i < m->triangles; i++) {
    if (tr[i] == 0) {
      for (j = m->triangles-1; j > i && tr[j] == 0; j--);
      if (j > i) {
	tr[j] = 0;
	tr[i] = 1;
	tl[i*3 + 0] = tl[j*3 + 0];
	tl[i*3 + 1] = tl[j*3 + 1];
	tl[i*3 + 2] = tl[j*3 + 2];
	m->textures[i] = m->textures[j];
	m->uv_list[i*6 + 0] = m->uv_list[j*6 + 0];
	m->uv_list[i*6 + 1] = m->uv_list[j*6 + 1];
	m->uv_list[i*6 + 2] = m->uv_list[j*6 + 2];
	m->uv_list[i*6 + 3] = m->uv_list[j*6 + 3];
	m->uv_list[i*6 + 4] = m->uv_list[j*6 + 4];
	m->uv_list[i*6 + 5] = m->uv_list[j*6 + 5];
      }
      else
	i = m->triangles;
    }
  }

  /* apply final changes */
  m->triangles = p1;
  m->vertices = p2;

  /* free the memory */
  free(ct);
  free(ve);
  free(tr);
  free(co);

  return SUCCEEDED;
}
