
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "../common/defines.h"
#include "../common/mesh.h"
#include "txt.h"
#include "simplify.h"


/*
#define _SIMPLIFY_DEBUG 1
#define _INTEGER_DEBUG 1
*/


#define COST_MAX 10000.0


/* NOTE! i'm using a simplification algorithm my stan melax
   http://www.melax.com/polychop/index.html */


struct vertex *vertices;



float _min(float a, float b) {

  if (a < b)
    return a;

  return b;
}


float _max(float a, float b) {

  if (a > b)
    return a;

  return b;
}


/* compute the cost for moving vertex a to b */
float _compute_cost(struct object_mesh *m, int a, int b, int *triangles, int t, int *faces, int f) {

  float cv, dx, dy, dz, e, mc, *fnl;
  int i, *tri, p1, p2, p3, j;


  if (f == 0 || t == 0) {
    fprintf(stderr, "_COMPUTE_COST: Uknown edge %d <-> %d!\n", a, b);
    exit(0);
  }

  /* none or far too many faces sharing this edge -> ultra high cost */
  if (f == 256 || t == 256)
    return COST_MAX;

  /* check that all the faces use the same texture, otherwise
     the cost will be ultra high! */
  i = m->textures[triangles[0]/3];
  for (j = 1; j < t; j++) {
    if (m->textures[triangles[j]/3] != i)
      return COST_MAX;
  }

  /* compute the cost */
  cv = 0;
  tri = m->triangle_list;
  fnl = m->face_normal_list;
  for (i = 0; i < t; i++) {
    p2 = triangles[i];
    mc = 2;
    for (j = 0; j < f; j++) {
      /* take the dot product of the face normals */
      for (p3 = 0; p3 < f; p3++) {
	if (p2 == faces[p3])
	  break;
      }
      if (p3 == f) {
	p1 = faces[j];
	dx = fnl[p2 + 0]*fnl[p1 + 0] + fnl[p2 + 1]*fnl[p1 + 1] + fnl[p2 + 2]*fnl[p1 + 2];
	mc = _min(mc, 1 - dx);
      }
    }
    cv = _max(cv, mc);
  }

  /* the compute edge length */
  a *= 3;
  b *= 3;
  dx = m->vertice_list[a + 0] - m->vertice_list[b + 0];
  dy = m->vertice_list[a + 1] - m->vertice_list[b + 1];
  dz = m->vertice_list[a + 2] - m->vertice_list[b + 2];
  e = sqrt(dx*dx + dy*dy + dz*dz);

  return e*cv/2;
}


int _integer_replace(int *l, int v, int r, int s) {

  int i;


#ifdef _INTEGER_DEBUG
  fprintf(stderr, "_INTEGER_REPLACE: %d with %d OLD: ", v/3, r/3);
  for (i = 0; i < s; i++)
    fprintf(stderr, "%d ", l[i]/3);
  fprintf(stderr, "\n");
#endif

  for (i = 0; i < s; i++) {
    if (l[i] == v) {
      l[i] = r;
      return SUCCEEDED;
    }
  }

  return FAILED;
}


int _integer_find(int *l, int v, int s) {

  int i;


  for (i = 0; i < s; i++) {
    if (l[i] == v)
      return SUCCEEDED;
  }

  return FAILED;
}


int _integer_remove(int *l, int v, int *s) {

  int i;


#ifdef _INTEGER_DEBUG
  fprintf(stderr, "_INTEGER_REMOVE: %d OLD: ", v/3);
  for (i = 0; i < *s; i++)
    fprintf(stderr, "%d ", l[i]/3);
  fprintf(stderr, "\n");
#endif

  for (i = 0; i < *s; i++) {
    if (l[i] == v)
      break;
  }

  /* v is not in l */
  if (i == *s)
    return SUCCEEDED;

  *s = *s - 1;

  /* is v at the end of l? */
  if (i == *s) {
    return SUCCEEDED;
  }

  /* take one away from the end */
  l[i] = l[*s];

  return SUCCEEDED;
}


int _integer_add(int **l, int v, int *s, int *ms) {

  int i, *t;


  t = *l;

#ifdef _INTEGER_DEBUG
  fprintf(stderr, "_INTEGER_ADD: %d OLD: ", v/3);
  for (i = 0; i < *s; i++)
    fprintf(stderr, "%d ", t[i]/3);
  fprintf(stderr, "\n");
#endif

  for (i = 0; i < *s; i++) {
    if (t[i] == v)
      return SUCCEEDED;
  }

  *s = *s + 1;
  if (*s > *ms) {
    *ms = *ms + 4;
    t = realloc(*l, *ms * sizeof(int));
    if (t == NULL) {
      fprintf(stderr, "_INTEGER_ADD: Out of memory error.\n");
      return FAILED;
    }
    *l = t;
  }

  t[*s - 1] = v;

  return SUCCEEDED;
}


int _add_neighbour(int ver1, int ver2, int tri) {

  struct vertex *v1, *v2;


  if (ver1 == ver2) {
    fprintf(stderr, "_ADD_NEIGHBOUR: %d and %d are the same vertex!\n", ver1, ver2);
    return FAILED;
  }

  v1 = &vertices[ver1];
  v2 = &vertices[ver2];

  ver1 *= 3;
  ver2 *= 3;
  tri *= 3;

  /* add v1 to v2's list */
  _integer_add(&(v2->vertices), ver1, &(v2->nvertices), &(v2->mvertices));
  _integer_add(&(v2->triangles), tri, &(v2->ntriangles), &(v2->mtriangles));

  /* add v2 to v1's list */
  _integer_add(&(v1->vertices), ver2, &(v1->nvertices), &(v1->mvertices));
  _integer_add(&(v1->triangles), tri, &(v1->ntriangles), &(v1->mtriangles));

  return SUCCEEDED;
}


int _list_edge_triangles(struct object_mesh *m, int a, int b, int *out, int *s) {

  struct vertex *v;
  int i, t, *l, p1, p2, p3, o;


  v = &vertices[a];
  l = m->triangle_list;
  o = 0;
  for (i = 0; i < v->ntriangles && o < 512; i++) {
    t = v->triangles[i];
    p1 = l[t + 0];
    p2 = l[t + 1];
    p3 = l[t + 2];
    if ((p1 == a || p2 == a || p3 == a) && (p1 == b || p2 == b || p3 == b)) {
      out[o] = t;
      o++;
    }
  }

  if (o == 512) {
    fprintf(stderr, "_LIST_EDGE_TRIANGLES: Overflow!\n");
    exit(0);
  }

  *s = o;

  return SUCCEEDED;
}


int simplify_mesh(struct object_mesh *m, float quality) {

  struct vertex *vj, *vk, *vl;
  float *ct, min, t;
  char *ve, *tr, *co;
  int i, j, k, l, mv, *tl, ri, rj, p1, p2, p3, *in;
  int fjk, fjl, fkl, facesjk[512], facesjl[512], faceskl[512];
  double *vel;


  if (m == NULL)
    return FAILED;

  mv = m->vertices;
  ct = malloc(mv * sizeof(float));
  in = malloc(mv * sizeof(int));
  ve = malloc(mv);
  co = malloc(m->triangles);
  tr = malloc(m->triangles);
  vertices = malloc(mv * sizeof(struct vertex));

  if (ct == NULL || ve == NULL || tr == NULL || in == NULL || co == NULL || vertices == NULL) {
    if (ct != NULL)
      free(ct);
    if (ve != NULL)
      free(ve);
    if (tr != NULL)
      free(tr);
    if (in != NULL)
      free(in);
    if (co != NULL)
      free(co);
    if (vertices != NULL)
      free(vertices);
    fprintf(stderr, "SIMPLIFY_MESH: Out of memory error.\n");
    return FAILED;
  }

  for (i = 0; i < mv; i++) {
    ve[i] = 0;
    vertices[i].nvertices = 0;
    vertices[i].mvertices = 0;
    vertices[i].ntriangles = 0;
    vertices[i].mtriangles = 0;
    vertices[i].vertices = NULL;
    vertices[i].triangles = NULL;
  }

  /* init to: keep all triangles */
  for (i = 0; i < m->triangles; i++)
    tr[i] = 1;

  /* init to: keep only used vertices */
  for (i = 0; i < m->triangles*3; i++) {
    ve[m->triangle_list[i]] = 1;
  }

  /* init to: maximum cost for every move */
  for (i = 0; i < mv; i++) {
    ct[i] = COST_MAX;
    in[i] = 0;
  }

  /* init to: no computations done */
  for (i = 0; i < m->triangles; i++)
    co[i] = 0;

  /* create the neighbour network */
  tl = m->triangle_list;
  for (i = 0; i < m->triangles; i++, tl += 3) {
    _add_neighbour(tl[0], tl[1], i);
    _add_neighbour(tl[0], tl[2], i);
    _add_neighbour(tl[1], tl[2], i);
  }

  /* DEBUG
  for (i = 0; i < mv; i++) {
    fprintf(stderr, "%d: %d neighbours: ", i, vertices[i].nvertices);
    for (j = 0; j < vertices[i].nvertices; j++)
      fprintf(stderr, "%d ", vertices[i].vertices[j]);
    fprintf(stderr, "--- %d triangles: ", vertices[i].ntriangles);
    for (j = 0; j < vertices[i].ntriangles; j++)
      fprintf(stderr, "%d ", vertices[i].triangles[j]);
    fprintf(stderr, "\n");
  }
  */

  /* find loose vertices */
  for (i = 0; i < mv; i++) {
    if (ve[i] == 0)
      fprintf(stderr, "SIMPLIFY_MESH: Loose vertex %d!\n", i);
  }

  /* check no vertex knows itself */
  for (i = 0; i < mv; i++) {
    vj = &vertices[i];
    if (_integer_find(vj->vertices, i*3, vj->nvertices) == SUCCEEDED)
      fprintf(stderr, "SIMPLIFY_MESH: Vertex %d knows itself!\n", i);
  }

  /* collapse edges until the quality threshold is reached */
  tl = m->triangle_list;
  while (1) {

    /* find costs for every move */
    for (i = 0; i < m->triangles; i++) {
      if (tr[i] == 1 && co[i] == 0) {
	j = tl[i*3 + 0];
	k = tl[i*3 + 1];
	l = tl[i*3 + 2];

	if (ve[j] == 0 || ve[k] == 0 || ve[l] == 0) {
	  fprintf(stderr, "SIMPLIFY_MESH: (%d %d %d) vertices (%d %d %d)!\n", ve[j], ve[k], ve[l], j, k, l);
	  exit(0);
	}

	/* find the faces using these two vertices */
	_list_edge_triangles(m, j, k, facesjk, &fjk);
	_list_edge_triangles(m, j, l, facesjl, &fjl);
	_list_edge_triangles(m, k, l, faceskl, &fkl);

	vj = &vertices[j];
	vk = &vertices[k];
	vl = &vertices[l];

	/* move vertex 0 to 1 and 2 */
	t = _compute_cost(m, j, k, vj->triangles, vj->ntriangles, facesjk, fjk);
	if (ct[j]> t) {
	  ct[j] = t;
	  in[j] = k;
	}
	t = _compute_cost(m, j, l, vj->triangles, vj->ntriangles, facesjl, fjl);
	if (ct[j] > t) {
	  ct[j] = t;
	  in[j] = l;
	}

	/* move vertex 1 to 0 and 2 */
	t = _compute_cost(m, k, j, vk->triangles, vk->ntriangles, facesjk, fjk);
	if (ct[k] > t) {
	  ct[k] = t;
	  in[k] = j;
	}
	t = _compute_cost(m, k, l, vk->triangles, vk->ntriangles, faceskl, fkl);
	if (ct[k] > t) {
	  ct[k] = t;
	  in[k] = l;
	}

	/* move vertex 2 to 0 and 1 */
	t = _compute_cost(m, l, j, vl->triangles, vl->ntriangles, facesjl, fjl);
	if (ct[l] > t) {
	  ct[l] = t;
	  in[l] = j;
	}
	t = _compute_cost(m, l, k, vl->triangles, vl->ntriangles, faceskl, fkl);
	if (ct[l] > t) {
	  ct[l] = t;
	  in[l] = k;
	}
	/* mark it computed */
	co[i] = 1;
      }
    }

    /* find the minimum cost */
    min = COST_MAX;
    ri = 0;
    rj = 0;
    for (i = 0; i < mv; i++) {
      if (ve[i] == 1 && ve[in[i]] == 1 && ct[i] < min) {
	ri = i;
	rj = in[i];
	min = ct[i];
      }
    }

    if (ri == rj)
      break;

    /* have we reached the limit? */
    if (min > quality)
      break;

#ifdef _SIMPLIFY_DEBUG
    fprintf(stderr, "SIMPLIFY_MESH: Move %d->%d cost %f.\n", ri, rj, min);
#endif

    /* remove the faces that use these two vertices */
    _list_edge_triangles(m, ri, rj, facesjk, &fjk);

    if (fjk == 0) {
      fprintf(stderr, "SIMPLIFY_MESH: Internal network is corrupted (1) %d %d!\n", ri, rj);
      fprintf(stderr, "VERTEX: %d(%d)\n", ri, ve[ri]);
      vj = &vertices[ri];
      fprintf(stderr, "VERTICES: ");
      for (i = 0; i < vj->nvertices; i++)
	fprintf(stderr, "%d(%d) ", vj->vertices[i]/3, ve[vj->vertices[i]/3]);
      fprintf(stderr, "\n");
      fprintf(stderr, "TRIANGLES: ");
      for (i = 0; i < vj->ntriangles; i++)
	fprintf(stderr, "%d(%d) ", vj->triangles[i]/3, tr[vj->triangles[i]/3]);
      fprintf(stderr, "\n");
      fprintf(stderr, "VERTEX: %d(%d)\n", rj, ve[rj]);
      vj = &vertices[rj];
      fprintf(stderr, "VERTICES: ");
      for (i = 0; i < vj->nvertices; i++)
	fprintf(stderr, "%d(%d) ", vj->vertices[i]/3, ve[vj->vertices[i]/3]);
      fprintf(stderr, "\n");
      fprintf(stderr, "TRIANGLES: ");
      for (i = 0; i < vj->ntriangles; i++)
	fprintf(stderr, "%d(%d) ", vj->triangles[i]/3, tr[vj->triangles[i]/3]);
      fprintf(stderr, "\n");
      exit(0);
    }

    /* find the destination texture coordinates */
    j = facesjk[0];
    if (tl[j + 0] == rj)
      fkl = 0;
    else if (tl[j + 1] == rj)
      fkl = 2;
    else if (tl[j + 2] == rj)
      fkl = 4;
    else {
      fprintf(stderr, "SIMPLIFY_MESH: Internal network is corrupted (2)!\n");
      exit(0);
    }

    vj = &vertices[ri];
    vk = &vertices[rj];

    /* check if we destroyed a lone triangle */
    if (vj->ntriangles == 1) {
      j = vj->triangles[0];

#ifdef _SIMPLIFY_DEBUG
      fprintf(stderr, "SIMPLIFY_MESH: LONE TRIANGLE DESTROYED: %d\n", j/3);
#endif

      tr[j/3] = 0;
      co[j/3] = 1;
      p1 = tl[j + 0];
      p2 = tl[j + 1];
      p3 = tl[j + 2];
      if (p1 != ri && p1 != rj)
	k = p1;
      else if (p2 != ri && p2 != rj)
	k = p2;
      else
	k = p3;

#ifdef _SIMPLIFY_DEBUG
      fprintf(stderr, "SIMPLIFY_MESH: TAKING OUT VERTICES %d and %d\n", k, ri);
#endif

      ve[k] = 0;
      ct[k] = COST_MAX;
      ve[ri] = 0;
      ct[ri] = COST_MAX;
      continue;
    }

    for (i = 0; i < fjk; i++) {
      j = facesjk[i];
      if (tr[j/3] == 0) {
	fprintf(stderr, "SIMPLIFY_MESH: Deleting a nonexisting triangle %d!\n", j/3);
      }
      tr[j/3] = 0;
      p1 = tl[j + 0];
      p2 = tl[j + 1];
      p3 = tl[j + 2];
      vj = &vertices[p1];
      vk = &vertices[p2];
      vl = &vertices[p3];

#ifdef _SIMPLIFY_DEBUG
      fprintf(stderr, "SIMPLIFY_MESH: Deleting triangle %d\n", j/3);
#endif

      _integer_remove(vj->triangles, j, &(vj->ntriangles));
      _integer_remove(vk->triangles, j, &(vk->ntriangles));
      _integer_remove(vl->triangles, j, &(vl->ntriangles));
      _integer_remove(vj->vertices, rj*3, &(vj->nvertices));
      _integer_remove(vk->vertices, rj*3, &(vk->nvertices));
      _integer_remove(vl->vertices, rj*3, &(vl->nvertices));
    }

    vj = &vertices[ri];
    vk = &vertices[rj];

    /* remove ri from rj, and rj from ri */
    _integer_remove(vk->vertices, ri*3, &(vk->nvertices));
    _integer_remove(vj->vertices, rj*3, &(vj->nvertices));

    /* add ri's triangles and vertices to rj */
    for (k = 0; k < vj->nvertices; k++) {
      _integer_add(&(vk->vertices), vj->vertices[k], &(vk->nvertices), &(vk->mvertices));
    }
    for (k = 0; k < vj->ntriangles; k++) {
      _integer_add(&(vk->triangles), vj->triangles[k], &(vk->ntriangles), &(vk->mtriangles));
    }

    /* take the vertex away */
    ve[ri] = 0;
    ct[ri] = COST_MAX;

    /* move it into the other one */
    for (k = 0; k < vj->nvertices; k++) {
      vl = &vertices[vj->vertices[k]/3];
      _integer_replace(vl->vertices, ri*3, rj*3, vl->nvertices);
    }

    vel = m->vertice_list;
    tl = m->triangle_list;
    for (k = 0; k < vj->ntriangles; k++) {
      i = vj->triangles[k];

#ifdef _SIMPLIFY_DEBUG
      fprintf(stderr, "SIMPLIFY_MESH: OLD TRI: (%d %d %d)\n", tl[i+0], tl[i+1], tl[i+2]);
#endif

      if (tl[i + 0] == ri) {
	tl[i + 0] = rj;
	m->uv_list[i*2 + 0] = m->uv_list[facesjk[0]*2 + fkl + 0];
	m->uv_list[i*2 + 1] = m->uv_list[facesjk[0]*2 + fkl + 1];
      }
      else if (tl[i + 1] == ri) {
	tl[i + 1] = rj;
	m->uv_list[i*2 + 2] = m->uv_list[facesjk[0]*2 + fkl + 0];
	m->uv_list[i*2 + 3] = m->uv_list[facesjk[0]*2 + fkl + 1];
      }
      else if (tl[i + 2] == ri) {
	tl[i + 2] = rj;
	m->uv_list[i*2 + 4] = m->uv_list[facesjk[0]*2 + fkl + 0];
	m->uv_list[i*2 + 5] = m->uv_list[facesjk[0]*2 + fkl + 1];
      }
      else {
	fprintf(stderr, "SIMPLIFY_MESH: Internal network is corrupted (3)!\n");
	exit(0);
      }

#ifdef _SIMPLIFY_DEBUG
      fprintf(stderr, "SIMPLIFY_MESH: NEW TRI: (%d %d %d)\n", tl[i+0], tl[i+1], tl[i+2]);

      /*
      for (j = 0; j < vk->nvertices; j++) {
	vj = &vertices[vk->vertices[j]/3];
	fprintf(stderr, "VERTEX: %d\n", vk->vertices[j]/3);
	fprintf(stderr, "VERTICES: ");
	for (k = 0; k < vj->nvertices; k++)
	  fprintf(stderr, "%d(%d) ", vj->vertices[k]/3, ve[vj->vertices[k]/3]);
	fprintf(stderr, "\n");
	fprintf(stderr, "TRIANGLES: ");
	for (k = 0; k < vj->ntriangles; k++)
	  fprintf(stderr, "%d(%d) ", vj->triangles[k]/3, tr[vj->triangles[k]/3]);
	fprintf(stderr, "\n");
	}
      */
#endif

      /* recompute the face normal and mark it uncomputed */
      {
	double x1, y1, z1, x2, y2, z2, x3, y3, z3, dx1, dy1, dz1, dx2, dy2, dz2;
	int n;

	n = tl[i + 0]*3;
	x1 = vel[n + 0];
	y1 = vel[n + 1];
	z1 = vel[n + 2];
	n = tl[i + 1]*3;
	x2 = vel[n + 0];
	y2 = vel[n + 1];
	z2 = vel[n + 2];
	n = tl[i + 2]*3;
	x3 = vel[n + 0];
	y3 = vel[n + 1];
	z3 = vel[n + 2];

	dx1 = x2 - x1;
	dy1 = y2 - y1;
	dz1 = z2 - z1;
	dx2 = x3 - x1;
	dy2 = y3 - y1;
	dz2 = z3 - z1;
	x1 = dy1*dz2 - dy2*dz1;
	x2 = dx2*dz1 - dx1*dz2;
	x3 = dx1*dy2 - dx2*dy1;

	dx1 = sqrt(x1*x1 + x2*x2 + x3*x3);

	m->face_normal_list[i + 0] = (float)(x1/dx1);
	m->face_normal_list[i + 1] = (float)(x2/dx1);
	m->face_normal_list[i + 2] = (float)(x3/dx1);

	co[i/3] = 0;
      }
    }

    /* remove all lone vertices */
    for (k = 0; k < vk->nvertices; k++) {
      i = vk->vertices[k]/3;
      vl = &vertices[i];
      if (vl->nvertices == 1) {
	j = vl->vertices[0]/3;
	vj = &vertices[j];
	ve[i] = 0;
	ct[i] = COST_MAX;
	_integer_remove(vj->vertices, i*3, &(vj->nvertices));
      }
    }
  }

  /* init to: keep no vertices */
  for (i = 0; i < m->vertices; i++)
    ve[i] = 0;

  /* keep only those vertices that the triangles use */
  for (i = 0; i < m->triangles; i++) {
    if (tr[i] == 1) {
      ve[tl[i*3 + 0]] = 1;
      ve[tl[i*3 + 1]] = 1;
      ve[tl[i*3 + 2]] = 1;
    }
  }

#ifdef _SIMPLIFY_DEBUG
  /* debug output */
  fprintf(stderr, "TRIANGLES:\n");
  for (i = 0; i < m->triangles; i++)
    fprintf(stderr, "%d ", tr[i]);
  fprintf(stderr, "\n");

  fprintf(stderr, "VERTICES:\n");
  for (i = 0; i < m->vertices; i++)
    fprintf(stderr, "%d ", ve[i]);
  fprintf(stderr, "\n");
#endif

  /* count the remaining triangles and vertices */
  for (i = 0, p1 = 0; i < m->triangles; i++) {
    if (tr[i] == 1)
      p1++;
  }

  for (i = 0, p2 = 0; i < m->vertices; i++) {
    if (ve[i] == 1)
      p2++;
  }

  fprintf(stderr, "SIMPLIFY_MESH: mesh \"%s\", triangles %d -> %d, vertices %d -> %d\n", m->name, m->triangles, p1, m->vertices, p2);

  /* compact the vertices */
  for (i = 0, j = m->vertices - 1; i < m->vertices; i++) {
    if (ve[i] == 0) {
      for ( ; j > i && ve[j] == 0; j--);
      if (j > i) {
	ve[j] = 0;
	ve[i] = 1;
	m->vertice_list[i*3 + 0] = m->vertice_list[j*3 + 0];
	m->vertice_list[i*3 + 1] = m->vertice_list[j*3 + 1];
	m->vertice_list[i*3 + 2] = m->vertice_list[j*3 + 2];
	/* update the triangles */
	for (ri = 0; ri < m->triangles*3; ri++) {
	  if (tl[ri] == j)
	    tl[ri] = i;
	}
	j--;
      }
      else
	i = m->vertices;
    }
  }

  /* compact triangles */
  for (i = 0, j = m->triangles - 1; i < m->triangles; i++) {
    if (tr[i] == 0) {
      for ( ; j > i && tr[j] == 0; j--);
      if (j > i) {
	tr[j] = 0;
	tr[i] = 1;
	tl[i*3 + 0] = tl[j*3 + 0];
	tl[i*3 + 1] = tl[j*3 + 1];
	tl[i*3 + 2] = tl[j*3 + 2];
	m->textures[i] = m->textures[j];
	m->uv_list[i*6 + 0] = m->uv_list[j*6 + 0];
	m->uv_list[i*6 + 1] = m->uv_list[j*6 + 1];
	m->uv_list[i*6 + 2] = m->uv_list[j*6 + 2];
	m->uv_list[i*6 + 3] = m->uv_list[j*6 + 3];
	m->uv_list[i*6 + 4] = m->uv_list[j*6 + 4];
	m->uv_list[i*6 + 5] = m->uv_list[j*6 + 5];
	j--;
      }
      else
	i = m->triangles;
    }
  }

  /* apply final changes */
  m->triangles = p1;
  m->vertices = p2;

  /* free the memory */
  free(ct);
  free(ve);
  free(tr);
  free(co);

  return SUCCEEDED;
}
