
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "main.h"
#include "id.h"
#include "psystems.h"
#include "../common/parse.h"


extern unsigned int current_time, *world_time;
extern int world_x, world_y, world_z;
extern char world_name[256];

struct psystem **psystem_map = NULL, **psystem_table = NULL;


int psystems_init(void) {

  int i;


  psystem_map = malloc(sizeof(struct psystem *) * world_x*world_y*world_z);
  psystem_table = malloc(sizeof(struct psystem *) * MAX_ID);
  if (psystem_map == NULL || psystem_table == NULL) {
    fprintf(stderr, "PSYSTEMS_INIT: Out of memory error.\n");
    return FAILED;
  }

  for (i = 0; i < world_x*world_y*world_z; i++)
    psystem_map[i] = NULL;

  for (i = 0; i < MAX_ID; i++)
    psystem_table[i] = NULL;

  return SUCCEEDED;
}


struct psystem *psystems_get(unsigned int id) {

  if (id < 0 || id >= MAX_ID) {
    fprintf(stderr, "PSYSTEMS_GET: ID %d out of range.\n", id);
    return NULL;
  }

  if (psystem_table[id] == NULL)
    fprintf(stderr, "PSYSTEMS_GET: Missing object %d.\n", id);

  return psystem_table[id];
}


void psystems_remove(struct psystem *g) {

  char tmp[8];
  int i;


  if (g->id < 0 || g->id > MAX_ID) {
    fprintf(stderr, "PSYSTEMS_REMOVE: ID %d is out of range.\n", g->id);
    return;
  }

  if (g->prev == NULL) {
    i = g->cz*world_y*world_x + g->cy*world_x + g->cx;
    if (psystem_map[i] != g) {
      fprintf(stderr, "PSYSTEMS_REMOVE: Mangled object %d.\n", g->id);
      return;
    }
    psystem_map[i] = g->next;
  }

  psystem_table[g->id] = NULL;

  if (g->next != NULL)
    g->next->prev = g->prev;
  if (g->prev != NULL)
    g->prev->next = g->next;

  /* remove message */
  tmp[0] = 'X';
  write_int(tmp + 1, g->id);

  send_to_all(tmp, 5);
  free_id(g->id);
  free(g);
}


void psystems_add(struct psystem *g) {

  int x, y, z, i;


  if (g->id < 0 || g->id > MAX_ID) {
    fprintf(stderr, "PSYSTEM_ADD: ID %d is out of range.\n", g->id);
    return;
  }

  if (psystem_table[g->id] != NULL)
    fprintf(stderr, "PSYSTEM_ADD: Duplicate ID %d.\n", g->id);
  psystem_table[g->id] = g;

  x = g->x / CUBE_SIDE;
  y = g->y / CUBE_SIDE;
  z = g->z / CUBE_SIDE;

  if (x < 0 || x >= world_x)
    x = 0;
  if (y < 0 || y >= world_y)
    y = 0;
  if (z < 0 || z >= world_z)
    z = 0;

  g->cx = x;
  g->cy = y;
  g->cz = z;

  i = z*world_y*world_x + y*world_x + x;

  g->next = psystem_map[i];
  g->prev = NULL;
  if (psystem_map[i] != NULL)
    psystem_map[i]->prev = g;
  psystem_map[i] = g;

  /* mark the voxel as "dirty" */
  world_time[i] = current_time;
  g->change_time = current_time;
}


int psystems_load_instances(void) {

  unsigned int id;
  double x, y, z;
  float rx, ry, rz;
  FILE *f;
  int d, t;
  char a[256];


  strcpy(a, "../data/wld/");
  strcat(a, world_name);
  strcat(a, "/psystems.txt");

  f = fopen(a, "rb");
  if (f == NULL) {
    fprintf(stderr, "PSYSTEMS_LOAD_INSTANCES: Could not open \"%s\".\n", a);
    return FAILED;
  }

  d = 0;
  while (fscanf(f, "%d %lf %lf %lf %f %f %f", &t, &x, &y, &z, &rx, &ry, &rz) != EOF) {
    if (generate_id(&id) == FAILED) {
      fclose(f);
      return FAILED;
    }
    psystems_create(t, id, x, y, z, rx, ry, rz, NULL);
    d++;
  }

  fprintf(stderr, "PSYSTEMS_LOAD_INSTANCES: Loaded %d objects.\n", d);
  fclose(f);

  return SUCCEEDED;
}


int psystems_create(int type, unsigned int id, double x, double y, double z, float rx, float ry, float rz, struct psystem **o) {

  struct psystem *g;


  g = malloc(sizeof(struct psystem));
  if (g == NULL) {
    fprintf(stderr, "PSYSTEMS_CREATE: Out of memory error.\n");
    return FAILED;
  }

  g->id = id;
  g->type = type;
  g->x = x;
  g->y = y;
  g->z = z;
  g->rx = rx;
  g->ry = ry;
  g->rz = rz;

  psystems_add(g);

  if (o != NULL)
    *o = g;

  return SUCCEEDED;
}


void psystems_relocate(struct psystem *g) {

  unsigned int x, y, z, i;


  x = g->x / CUBE_SIDE;
  y = g->y / CUBE_SIDE;
  z = g->z / CUBE_SIDE;

  if (x < 0 || x >= world_x)
    x = 0;
  if (y < 0 || y >= world_y)
    y = 0;
  if (z < 0 || z >= world_z)
    z = 0;

  /* is the object still in the right cube? */
  if (g->cx == x && g->cy == y && g->cz == z)
    return;

  /* cube changed -> relocate the object */
  if (g->next != NULL)
    g->next->prev = g->prev;
  if (g->prev != NULL)
    g->prev->next = g->next;
  if (g->prev == NULL) {
    i = g->cz*world_y*world_x + g->cy*world_x + g->cx;
    psystem_map[i] = g->next;
  }

  g->cx = x;
  g->cy = y;
  g->cz = z;    

  i = z*world_y*world_x + y*world_x + x;
  g->prev = NULL;
  g->next = psystem_map[i];
  if (psystem_map[i] != NULL)
    psystem_map[i]->prev = g;
  psystem_map[i] = g;
}


